/*
 * Copyright (c) 2001 Massachusetts Institute of Technology.
 *
 * This file is part of bgptools. 
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `COPYING' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file AUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the AUTHORS file.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


/*
 *     Nick Feamster (feamster@lcs.mit.edu)
 *     MIT Lab for Computer Science
 *     November 2001
 */


#include "prefix_tree.h"

int
prefix_tree::find_depth(u_int32_t prefix) {
  int count = 0;
  while (!(prefix & 1)) {
    count++;
    prefix= prefix >> 1;
  }
  return count;
}

prefix_tree::prefix_tree()  {
  this->root = (prefix_node*)malloc(sizeof(prefix_node));
  this->root->prefix=0;
  this->root->left = NULL;
  this->root->right = NULL;
}

void
prefix_tree::free_tree(prefix_node *p) {
  if (p->left) { free(p->left); p->left=NULL; };
  if (p->right) { free(p->right); p->right=NULL; };
  if (p) { free(p); p=NULL; }
}

prefix_tree::~prefix_tree()  {
  free_tree(this->root);
}



void
prefix_tree::insert(char *filename) {

  FILE *fp;
  update upd;

  fp = fopen(filename, "rb");
  if (fp == NULL) { fprintf(stderr, "error: could not open %s\n", filename); };
  
  while(!feof(fp)) {
    read_record(fp, &upd);
    if (upd.ad_type == BGP_ADV) {
      this->insert(&upd);
    }
  }

}


void
prefix_tree::insert(struct update *upd) {

  struct prefix_s *prefixes = upd->prefix_;

  prefix_node *p = (prefix_node*)malloc(sizeof(prefix_node));
  p->prefix = prefixes->pfx;
  p->depth = prefixes->len;
#ifdef __PT_DEBUG__
  fprintf(stdout, "*** inserting %x, depth %d\n", prefixes->pfx, prefixes->len);
#endif
  this->insert(p, this->root, p->depth);
}


void
prefix_tree::insert(prefix_node *p, prefix_node *root, int depth) {
  u_int32_t prefix = p->prefix;
  prefix_node *cur = root;

  for (int i=32; i>=(33-depth); i--) {
//      int bit = ((prefix & ((int)pow(2,i)-1)))>>(i-1);
    int bit = (prefix>>(i-1) & 1);

    if (!bit && !cur->left)
      cur->left = new_node(prefix, 33-i, bit);
    else if (bit && !cur->right)
      cur->right = new_node(prefix, 33-i, bit);

//        FILE  *fd = fopen("/dev/fd/0", "w");
//        this->pdh(this->root, fd);
//        fclose(fd);

    if (cur->left && cur->right) {
      if(!(cur->left->prefix != cur->right->prefix)) {
#ifdef __PT_DEBUG__
	fprintf(stderr, "ERROR: at depth %d, children of 0x%x are both 0x%x\n", cur->depth,
		cur->prefix, cur->left->prefix);
#endif
	exit(0);
      }
    }
    
    cur = (bit?cur->right:cur->left);
  }
  
}


prefix_node*
prefix_tree::new_node(u_int32_t prefix, int depth, int bit) {

  prefix_node *tmp = (prefix_node*)malloc(sizeof(prefix_node));

  u_int32_t mask=0;
  for (int i=0; i< 32; i++) {
    mask = mask << 1;
    mask |= (i<depth);
  }

  tmp->prefix = prefix & mask;
#ifdef __PT_DEBUG__
  fprintf(stdout, "mask: 0x%x...", mask);
  fprintf(stdout, "inserting 0x%x (%d): 0x%x at depth %d\n", prefix, bit, tmp->prefix, depth);
#endif
  assert(((tmp->prefix >> (32-depth)) & 1) == bit);

  tmp->depth = depth;
  tmp->left = NULL;
  tmp->right = NULL;

  return tmp;


}




void
prefix_tree::print(prefix_node *p) {
  if (p != NULL) {
    fprintf(stdout, "prefix: 0x%x/%d\n", p->prefix, p->depth);
    this->print(p->left);
    this->print(p->right);
  }
}


void
prefix_tree::print_dot(prefix_node *p, char *filename) {

  this->node_num=0;
  FILE *fp = fopen(filename, "w");

  fprintf(fp, "digraph Prefix_Tree {\n");
  fprintf(fp, "node [shape=record,height=0.1];\n");
  this->pdh(p, fp);
  fprintf(fp, "}\n");

  fclose(fp);


#ifdef __USE_DOT__  
  char cmd[1024];
  char outfile[80];
  char tmp[80];

  strncpy(tmp, filename, strlen(filename));
  char *tok = strtok(tmp, ".");
  strncpy(outfile, tok, strlen(tmp));
  sprintf(outfile, "%s.ps", tok);
  outfile[strlen(tok)+3]='\0';

  fprintf(stderr, "printing to %s %s\n", filename, outfile);  
  sprintf(cmd, "%s/%s -Tps %s -o %s", DOT_DIR, "dot", filename, outfile);
  system(cmd);
#endif
}


void
prefix_tree::pdh(prefix_node *p, FILE *fp) {
  if (p != NULL) {
//      fprintf(fp, "node0x%x:%d [label = \" 0x%x:%d \"];\n", p->prefix, p->depth, p->prefix, p->depth);
    if (p->left != NULL) { fprintf(fp, " \"0x%x:%d\" -> \"0x%x:%d\"\n", p->prefix, p->depth, p->left->prefix, p->left->depth); }
    if (p->right != NULL) { fprintf(fp, " \"0x%x:%d\" -> \"0x%x:%d\"\n", p->prefix, p->depth, p->right->prefix,p->right->depth); }

    this->pdh(p->left, fp);
    this->pdh(p->right, fp);
  }
}




